self.__BUILD_MANIFEST=function(s,e,c,a,t){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},__routerFilterStatic:{numItems:0,errorRate:1e-4,numBits:0,numHashes:null,bitArray:[]},__routerFilterDynamic:{numItems:0,errorRate:1e-4,numBits:s,numHashes:null,bitArray:[]},"/":["static/chunks/pages/index-c5b4576aaa6e321d.js"],"/404":["static/chunks/pages/404-a3bc5648629410d5.js"],"/500":["static/chunks/pages/500-079d29db5bc05d18.js"],"/_error":["static/chunks/pages/_error-fde50cb7f1ab27e0.js"],"/article":[a,t,"static/chunks/pages/article-47973af8c72f6d61.js"],"/articles":["static/chunks/pages/articles-8cd425037154f5a6.js"],"/source":[a,t,"static/chunks/pages/source-5d3aabad73965638.js"],"/sources":["static/chunks/pages/sources-9bb7e675cd88cc75.js"],sortedPages:["/","/404","/500","/_app","/_error","/article","/articles","/source","/sources"]}}(0,0,0,"static/chunks/9-cf74b2a6aa463331.js","static/css/7664c1449da0a98b.css"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();