:root{--background:#fff;--foreground:#171717;--highlight:#f0f0f0;--border-color:#a0a0a0;--passive-text:#707070;--medium-border-radius:5px;--small-border-radius:5px;--margin-top:1rem;--margin:2rem}body,html{font-size:14pt}body{color:var(--foreground);background:var(--background);font-family:Times New Roman,Times,serif}*{box-sizing:border-box;padding:0;margin:0}