.MarkdownViewer_article__CXqGO h1,.MarkdownViewer_article__CXqGO h2,.MarkdownViewer_article__CXqGO h3,.MarkdownViewer_article__CXqGO ol,.MarkdownViewer_article__CXqGO p,.MarkdownViewer_article__CXqGO ul{margin-top:20px;margin-bottom:20px;line-height:1.5rem}.MarkdownViewer_article__CXqGO>p:first-child{margin-top:0}.MarkdownViewer_article__CXqGO>p:last-child{margin-bottom:0}.MarkdownViewer_article__CXqGO ol,.MarkdownViewer_article__CXqGO ul,.MarkdownViewer_editorArticle__Qa97H ol,.MarkdownViewer_editorArticle__Qa97H ul{padding-left:1.5em}.MarkdownViewer_article__CXqGO li{margin-top:10px}.MarkdownViewer_article__CXqGO li p,.MarkdownViewer_article__CXqGO ol p,.MarkdownViewer_article__CXqGO ul p{display:inline}@keyframes Modal_fadeIn__Oj_FH{0%{opacity:0}to{opacity:1}}.Modal_overlay__2Sh3U{position:fixed;top:0;left:0;right:0;bottom:0;background-color:rgba(0,0,0,.5);z-index:1000;display:flex;align-items:center;justify-content:center;cursor:default;animation-name:Modal_fadeIn__Oj_FH;animation-duration:.5s;animation-fill-mode:forwards}.Modal_container__jKKR0{display:flex;flex-direction:column;max-height:80vh;position:relative;background-color:var(--background);border:1px solid var(--border-color);border-radius:var(--medium-border-radius);transition:width 1s linear;color:var(--foreground)}.Modal_closeModal__QVqd2{cursor:pointer}.Modal_header__Pfr2K{display:flex;font-size:.9rem;align-items:center;background-color:var(--highlight);padding:10px;border-top-left-radius:var(--medium-border-radius);border-top-right-radius:var(--medium-border-radius)}.Modal_title__a7vfd{flex:1 1;color:var(--passive-text)}.Modal_body__KZX7m{margin:10px;overflow-y:scroll}