(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[977],{8319:(e,t,s)=>{(window.__NEXT_P=window.__NEXT_P||[]).push(["/articles",function(){return s(5675)}])},5675:(e,t,s)=>{"use strict";s.r(t),s.d(t,{default:()=>c});var i=s(4848),l=s(6540),n=s(1106),r=s.n(n),a=s(6080);function c(){let[e,t]=(0,l.useState)([]),{setPath:s}=(0,a.ok)();return(0,l.useEffect)(()=>(s([{href:{pathname:"/articles"},title:"Articles"}]),()=>{s([])}),[]),(0,l.useEffect)(()=>{(async function(){try{let e=await fetch("".concat("","/api/index"),{method:"GET"});if(!e.ok)throw Error("Response was not ok");let s=await e.json();t(s.entities)}catch(e){console.log(e)}})()},[]),(0,i.jsx)("div",{style:{display:"flex",flexDirection:"column"},children:e.map(e=>e.is_written?(0,i.jsx)("div",{children:(0,i.jsx)(r(),{href:{pathname:"/article",query:{e:e.slug}},children:e.title})},e.slug):(0,i.jsx)("div",{children:(0,i.jsx)(r(),{style:{all:"unset",cursor:"default"},href:{pathname:"/article",query:{e:e.slug}},children:e.title})},e.slug))})}}},e=>{var t=t=>e(e.s=t);e.O(0,[636,593,792],()=>t(8319)),_N_E=e.O()}]);