(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[332],{7276:(e,r,s)=>{(window.__NEXT_P=window.__NEXT_P||[]).push(["/",function(){return s(7031)}])},7031:(e,r,s)=>{"use strict";s.r(r),s.d(r,{default:()=>l});var i=s(4848);s(6540);var n=s(1106),c=s.n(n);function l(){return(0,i.jsxs)("div",{style:{display:"flex",flexDirection:"column"},children:[(0,i.jsx)("div",{children:(0,i.jsx)(c(),{href:"/articles",children:"Articles"})}),(0,i.jsx)("div",{children:(0,i.jsx)(c(),{href:"/sources",children:"Sources"})})]})}}},e=>{var r=r=>e(e.s=r);e.O(0,[636,593,792],()=>r(7276)),_N_E=e.O()}]);