(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[278],{8525:(e,t,s)=>{(window.__NEXT_P=window.__NEXT_P||[]).push(["/sources",function(){return s(5314)}])},5314:(e,t,s)=>{"use strict";s.r(t),s.d(t,{default:()=>u});var c=s(4848),n=s(6540),o=s(1106),r=s.n(o),i=s(6080);function u(){let[e,t]=(0,n.useState)([]),{setPath:s}=(0,i.ok)();return(0,n.useEffect)(()=>(s([{href:{pathname:"/sources"},title:"Sources"}]),()=>{s([])}),[]),(0,n.useEffect)(()=>{(async function(){try{let e=await fetch("".concat("","/api/sources"),{method:"GET"});if(!e.ok)throw Error("Response was not ok");let s=await e.json();t(s.sources)}catch(e){console.log(e)}})()},[]),(0,c.jsx)("div",{style:{display:"flex",flexDirection:"column"},children:e.map(e=>(0,c.jsx)("div",{children:(0,c.jsx)(r(),{href:{pathname:"/source",query:{id:e.id}},children:e.title})},e.id))})}}},e=>{var t=t=>e(e.s=t);e.O(0,[636,593,792],()=>t(8525)),_N_E=e.O()}]);